__all__ = ['Tactigon_RT_Computing', 'Data_Preprocessor']

from ...middleware.utilities.Data_Preprocessor import Data_Preprocessor
from ...middleware.utilities.Tactigon_RT_Computing import Tactigon_RT_Computing


import json
import requests
import os
import pandas as pd


from ..models import UserData, DataCollectionConfig 


def create_dir(path):
    """
    Create directories if doesn't exists
    :param path: path of the directory
    :return: None
    """
    if not os.path.exists(path):
        os.makedirs(path)
        print("directory created")
    else:
        print("directory already exists")


def user_login(url, user_data_path):
    """
    This function allow user to login if user info already exists otherwise ask for registration
    :param url: server url
    :param user_data_path: path to the user info file
    :return: login status
    """
    if not os.path.exists(user_data_path):
        # update_registration(user_data_path)
        pass
    else:
        with open(user_data_path) as json_file:
            user_data = json.load(json_file)
        # sending login request to server
        res = requests.post(url + "login", data=json.dumps(user_data))
        res = json.loads(res.content)

        if res["status"] == "wrongpassword":
            print("Your password is worng, please enter again")
            update_local_password(user_data_path, user_data)
            return res
        elif res["status"] == "wronguser":
            print("your user name does not exists, please register")
            update_local_registration(user_data_path)
            return res
        elif res["status"] == "successful":
            print("Login successful")
            return res
        else:
            print("register again")
            update_local_registration(user_data_path)
            return res


def update_local_password(user_data_path, user_data):
    """
    This function allow user to change password if it is wrong
    :param user_data_path: path of the user info file
    :param user_data: current user data
    :return: None
    """
    print("Please enter correct password:")
    password = input()
    user_data["password"] = password
    with open(user_data_path, "w") as outfile:
        json.dump(user_data, outfile)


def update_local_registration(user_data_path):
    """
    this function update local file with user's login information
    :param user_data_path: path of the file where user info will be stored
    :return: None
    """
    print("Enter your first name")
    first_name = input()
    print("Enter your second name")
    second_name = input()
    print("Enter a password")
    password = input()
    print("Enter your preferred user name(all lower case)")
    user_name = input()
    user_data = {
        "first_name": str(first_name),
        "last_name": str(second_name),
        "password": str(password),
        "user_id": str(user_name).lower(),
    }
    with open(user_data_path, "w") as outfile:
        json.dump(user_data, outfile)


def load_info(user_data: UserData, data_collection_config: DataCollectionConfig):
    """
    This function load info files and update info dataframes with config file values
    :param config: config object of data collection
    :return: person, gesture, and session dataframes
    """
    # import json config

    # raw_data_path = path.join(CLIENT_PY_PATH, dc_config.get("RAW_DATA_PATH"))

    path_person = data_collection_config.raw_data_full_path + "info/person.csv"
    path_gesture = data_collection_config.raw_data_full_path + "info/gesture.csv"
    path_session = data_collection_config.raw_data_full_path + "info/session.csv"
    path_device = data_collection_config.raw_data_full_path + "info/device.csv"

    ## If info file doesn't exists then create new info file with data
    ## otherwise load the already existing info
    if not os.path.exists(path_person):
        person_data = pd.DataFrame(columns=["person_name"])
        person_data.to_csv(path_person)
    else:
        person_data = pd.read_csv(path_person, index_col=0)

    if not os.path.exists(path_gesture):
        gestures_data = pd.DataFrame(columns=["gesture_names"])
        gestures_data.to_csv(path_gesture)
    else:
        gestures_data = pd.read_csv(path_gesture, index_col=0)

    if not os.path.exists(path_session):
        session_data = pd.DataFrame(columns=["session_info"])
        session_data.to_csv(path_session)
    else:
        session_data = pd.read_csv(path_session, index_col=0)   

    if not os.path.exists(path_device):
        device_data = pd.DataFrame(columns=["dev_id"])
        device_data.to_csv(path_device)
    else:
        device_data = pd.read_csv(path_device, index_col=0)

    return person_data, gestures_data, session_data, device_data


def update_dataframe(df, column, value):
    """
    This function update dataframe given column name and value if that values doesn't exist
    :param df: dataframe
    :param column: name of the column
    :param value: list of values
    :return: updated dataframe
    """
    for v in value:
        if get_index(df, column, v) == "":
            df.loc[len(df)] = v
    return df


def update_dataframe_2D___unused(df, column_1, column_2, value_1, value_2):

    for idx in df[df[column_1] == value_1].index.tolist():
        if idx in df[df[column_2] == value_2].index.tolist():
            # already exists
            return df
    
    df.loc[len(df)] = [value_1,value_2]
    return df

        
def update_info(raw_data_path, person_data, gesture_data, session_data, device_data):
    """
    This function save dataframes as csv file
    :param raw_data_path: path of the raw data folder
    :param person_data: dataframe with person info
    :param gesture_data: dataframe with gesture info
    :param session_data: dataframe with session info
    :return: none
    """

    person_data.to_csv(raw_data_path + "info/person.csv")
    gesture_data.to_csv(raw_data_path + "info/gesture.csv")
    session_data.to_csv(raw_data_path + "info/session.csv")
    device_data.to_csv(raw_data_path + "info/device.csv")
    print("data saved as csv\n\n")


def get_index(df, column, data):
    """
    Function returns the index of the row for a column value
    :param df: dataframe
    :param column: name of the column
    :param data: value you are searching
    :return: index of the row
    """
    idx = df[df[column] == data].index.tolist()
    idx = str(idx).replace("[", "")
    idx = idx.replace("]", "")
    return idx


def get_value(df, column, row_idx):
    """
    Function returns the value @ column/row_idx
    :param df: dataframe
    :param column: name of the column
    :row_idx: index of the row (starting from 0)
    :return: value @ column/row_idx
    """
    try:
        return df[column][row_idx]
    except:
        return None



from uuid import uuid4
from queue import Full, Empty
from multiprocessing import get_context
from multiprocessing.queues import Queue

from typing import Optional

class RollingQueue(Queue):
    def __init__(self, maxsize: int = 0):
        ctx = get_context()
        Queue.__init__(self, maxsize=maxsize, ctx=ctx)

    def put(self, obj, block: bool = True, timeout: Optional[float] = None):
        try:
            Queue.put(self, obj, block, timeout)
        except Full as e:
            Queue.get_nowait(self)
            Queue.put(self, obj, block, timeout)

class AdvancedQueue(Queue):
    _item_list = []

    def __init__(self, maxsize: int = 0):
        ctx = get_context()
        Queue.__init__(self, maxsize=maxsize, ctx=ctx)

    def put(self, item: any, block: bool = True, timeout: Optional[float] = None) -> str: #type: ignore
        _uuid = uuid4().hex
        Queue.put(self, (_uuid, item), block, timeout)
        self._item_list.append(_uuid)
        return _uuid

    def get(self, block: bool = True, timeout: Optional[float] = None):
        _uuid, item = Queue.get(self, block, timeout)
        self._item_list.pop(self._item_list.index(_uuid))
        return item

    def check(self, item: str) -> bool:
        return item not in self._item_list


import operator
import struct
from enum import Enum
from dataclasses import dataclass
from datetime import datetime
from functools import reduce

from typing import Optional, List, Tuple
    
class BLECommands(Enum):
    WRITE = 1
    READ = 2

class BLECommandWord(Enum):
    UPDATE_TOUCH1       = 0x41
    UPDATE_TOUCH2       = 0x42
    WRITE_TOUCH         = 0x43
    START               = 0x00
    STOP                = 0x01
    STATUS              = 0xFF

class BLECommandMask(Enum):
    SENSORFUSION        = b'\x00\x00\x01\x00'
    ECOMPASS            = b'\x00\x00\x00\x40'


class Hand(Enum):
    RIGHT = "right"
    LEFT = "left"

class TBleConnectionStatus(Enum):
    NONE = 0
    CONNECTING = 1
    CONNECTED = 2
    DISCONNECTING = 3
    DISCONNECTED = 4

class TBleSelector(Enum):
    NONE = 0
    SENSORS = 1
    AUDIO = 2

class OneFingerGesture(Enum):
    NONE = 0x00
    SINGLE_TAP = 0x01
    TAP_AND_HOLD = 0x02
    SWIPE_X_NEG = 0x04
    SWIPE_X_POS = 0x08
    SWIPE_Y_NEG = 0x20
    SWIPE_Y_POS = 0x10

class TwoFingerGesture(Enum):
    NONE = 0x00
    TWO_FINGER_TAP = 0x01
    SCROLL = 0x02
    ZOOM = 0x04

@dataclass
class BLECommand:
    command: BLECommands
    characteristic: str
    payload: Optional[bytes] = None

@dataclass
class Gesture:
    gesture: str
    probability: float
    confidence: float
    displacement: float

    def toJSON(self) -> dict:
        return {
            "gesture": self.gesture,
            "probability": self.probability,
            "confidence": self.confidence,
            "displacement": self.displacement,
        }

@dataclass
class Acceleration:
    x: float
    y: float
    z: float

    def toJSON(self) -> dict:
        return {
            "x": self.x,
            "y": self.y,
            "z": self.z,
        }

@dataclass
class Angle:
    roll: float
    pitch: float
    yaw: float

    def toJSON(self) -> dict:
        return {
            "roll": self.roll,
            "pitch": self.pitch,
            "yaw": self.yaw,
        }

@dataclass
class Gyro:
    x: float
    y: float
    z: float

    def toJSON(self) -> dict:
        return {
            "x": self.x,
            "y": self.y,
            "z": self.z,
        }

@dataclass
class Touch:
    one_finger: OneFingerGesture
    two_finger: TwoFingerGesture
    x_pos: float
    y_pos: float

    def __eq__(self, __value: object) -> bool:
        if isinstance(__value, Touch):
            return (
                self.one_finger == __value.one_finger 
                and self.two_finger == __value.two_finger
                and self.x_pos == __value.x_pos
                and self.y_pos == __value.y_pos
            )
        
        return False
    
    def toJSON(self) -> dict:
        return {
            "one_finger": self.one_finger.value,
            "two_finger": self.two_finger.value,
            "x_pos": self.x_pos,
            "y_pos": self.y_pos
        }

@dataclass
class TSkinState:
    connected: bool
    sleep: bool
    battery: Optional[float]
    selector: Optional[TBleSelector]
    touch: Optional[Touch]
    angle: Optional[Angle]
    gesture: Optional[Gesture]

    def toJSON(self) -> dict:
        return {
            "connected": self.connected,
            "sleep": self.sleep,
            "battery": self.battery,
            "selector": self.selector.value if self.selector else None,
            "touch": self.touch.toJSON() if self.touch else None,
            "angle": self.angle.toJSON() if self.angle else None,
            "gesture": self.gesture.toJSON() if self.gesture else None,
        }

@dataclass
class GestureConfig:
    model_path: str
    encoder_path: str
    name: str
    created_at: datetime
    gestures: Optional[List[str]] = None
    num_sample: int = 10
    gesture_prob_th: float = 0.85
    confidence_th: float = 5

    @classmethod
    def FromJSON(cls, json: dict):
        return cls(
            json["model_path"],
            json["encoder_path"],
            json["name"],
            datetime.fromisoformat(json["created_at"]),
            json["gestures"] if "gestures" in json else None,
            json["num_sample"] if "num_sample" in json and json["num_sample"] is not None else cls.num_sample,
            json["gesture_prob_th"] if "gesture_prob_th" in json and json["gesture_prob_th"] is not None else cls.gesture_prob_th,
            json["confidence_th"] if "confidence_th" in json and json["confidence_th"] is not None else cls.confidence_th,
            )
    
    def toJSON(self) -> dict:
        return {
            "model_path": self.model_path,
            "encoder_path": self.encoder_path,
            "name": self.name,
            "created_at": self.created_at.isoformat(),
            "gestures": self.gestures
        }

@dataclass
class TouchConfig:
    FORMAT_CFG1 = ">BBHHHHBHBHHH"
    FORMAT_CFG2 = ">BBHHHBBHHHHH"
    FORMAT_CMD  = ">BBHHHHHHHHH"

    swipe_initial_time: int
    swipe_initial_distance: int
    swipe_consecutive_time: int
    swipe_consecutive_distance: int
    swipe_angle: int
    scroll_initial_distance: int
    scroll_angle: int
    zoom_initial_distance: int
    zoom_consecutive_distance: int
    tap_time: int
    tap_distance: int
    hold_time: int
    one_finger_gesture: List[OneFingerGesture]
    two_finger_gesture: List[TwoFingerGesture]

    @classmethod
    def FromJSON(cls, json):
        return cls(
            json["swipe_initial_time"],
            json["swipe_initial_distance"],
            json["swipe_consecutive_time"],
            json["swipe_consecutive_distance"],
            json["swipe_angle"],
            json["scroll_initial_distance"],
            json["scroll_angle"],
            json["zoom_initial_distance"],
            json["zoom_consecutive_distance"],
            json["tap_time"],
            json["tap_distance"],
            json["hold_time"],
            [OneFingerGesture(f) for f in json["one_finger_gesture"]],
            [TwoFingerGesture(f) for f in json["two_finger_gesture"]],
        )
    
    @classmethod
    def Default(cls):
        return cls(
            swipe_initial_time=200,
            swipe_initial_distance=25,
            swipe_consecutive_time=150,
            swipe_consecutive_distance=25,
            swipe_angle=23,
            scroll_initial_distance=50,
            scroll_angle=37,
            zoom_initial_distance=50,
            zoom_consecutive_distance=5,
            tap_time=300,
            tap_distance=75,
            hold_time=10,
            one_finger_gesture=[OneFingerGesture.SINGLE_TAP, OneFingerGesture.TAP_AND_HOLD],
            two_finger_gesture=[TwoFingerGesture.TWO_FINGER_TAP]
        )
    
    @classmethod
    def High(cls):
        c = cls.Default()
        c.tap_time = 150
        c.tap_distance = 50
        return c

    @classmethod
    def Medium(cls):
        return cls.Default()

    @classmethod
    def Low(cls):
        c = cls.Default()
        c.tap_time = 600
        c.tap_distance = 150
        c.hold_time = 50
        return c

    def toJSON(self):
        return {
            "swipe_initial_time": self.swipe_initial_time,
            "swipe_initial_distance": self.swipe_initial_distance,
            "swipe_consecutive_time": self.swipe_consecutive_time,
            "swipe_consecutive_distance": self.swipe_consecutive_distance,
            "swipe_angle": self.swipe_angle,
            "scroll_initial_distance": self.scroll_initial_distance,
            "scroll_angle": self.scroll_angle,
            "zoom_initial_distance": self.zoom_initial_distance,
            "zoom_consecutive_distance": self.zoom_consecutive_distance,
            "tap_time": self.tap_time,
            "tap_distance": self.tap_distance,
            "hold_time": self.hold_time,
            "one_finger_gesture": [ofg.value for ofg in self.one_finger_gesture],
            "two_finger_gesture": [tfg.value for tfg in self.two_finger_gesture],
        }

    def set_sensitivity(self, sensitivity: int):
        if sensitivity == 1:
            self.tap_time = 600
            self.tap_distance = 150
            self.hold_time = 50
        elif sensitivity == 2:
            self.tap_time = 300
            self.tap_distance = 75
        else:
            self.tap_time = 150
            self.tap_distance = 50

    def toBytes(self) -> Tuple[bytes, bytes, bytes]:
        cfg1 = struct.pack(self.FORMAT_CFG1,
            BLECommandWord.UPDATE_TOUCH1.value,
            16,
            self.swipe_initial_time,
            self.swipe_initial_distance,
            self.swipe_consecutive_time,
            self.swipe_consecutive_distance,
            self.swipe_angle,
            self.scroll_initial_distance,
            self.scroll_angle,
            self.zoom_initial_distance,
            self.zoom_consecutive_distance,
            0
        )

        cfg2 = struct.pack(self.FORMAT_CFG2,
            BLECommandWord.UPDATE_TOUCH2.value,
            8,
            self.tap_time,
            self.tap_distance,
            self.hold_time,
            reduce(operator.ior, [g.value for g in self.one_finger_gesture]) if self.one_finger_gesture else 0,
            reduce(operator.ior, [g.value for g in self.two_finger_gesture]) if self.two_finger_gesture else 0,
            0,
            0,
            0,
            0,
            0,
        )

        save = struct.pack(self.FORMAT_CMD, BLECommandWord.WRITE_TOUCH.value,1,0,0,0,0,0,0,0,0,0)
        return cfg1, cfg2, save

    def update(self, cfg1: bytes, cfg2: bytes):
        word_cfg1, length_cfg1, *data_cfg1 = struct.unpack(self.FORMAT_CFG1, cfg1)
        word_cfg2, length_cfg2, *data_cfg2 = struct.unpack(self.FORMAT_CFG2, cfg2)

        if BLECommandWord(word_cfg1) != BLECommandWord.UPDATE_TOUCH1 or \
            BLECommandWord(word_cfg2) != BLECommandWord.UPDATE_TOUCH2 or \
            length_cfg1 != 16 or \
            length_cfg2 != 8:
            return False

        self.swipe_initial_time = data_cfg1[0]
        self.swipe_initial_distance = data_cfg1[1]
        self.swipe_consecutive_time = data_cfg1[2]
        self.swipe_consecutive_distance = data_cfg1[3]
        self.swipe_angle = data_cfg1[4]
        self.scroll_initial_distance = data_cfg1[5]
        self.scroll_angle = data_cfg1[6]
        self.zoom_initial_distance = data_cfg1[7]
        self.zoom_consecutive_distance = data_cfg1[8]
        self.tap_time = data_cfg2[0]
        self.tap_distance = data_cfg2[1]
        self.hold_time = data_cfg2[2]
        self.one_finger_gesture = [g for g in OneFingerGesture if bool(g.value & data_cfg2[3])]
        self.two_finger_gesture = [g for g in TwoFingerGesture if bool(g.value & data_cfg2[4])]

        return True
     

@dataclass
class TSkinConfig:
    address: str
    hand: Hand
    name: str = "Tactigon"
    touch_config: Optional[TouchConfig] = None
    gesture_config: Optional[GestureConfig] = None

    @classmethod
    def FromJSON(cls, json: dict):
        return cls(
            json["address"], 
            Hand(json["hand"]),
            json["name"] if "name" in json else cls.name,
            TouchConfig.FromJSON(json["touch_config"]) if "touch_config" in json and json["touch_config"] is not None else None,
            GestureConfig.FromJSON(json["gesture_config"]) if "gesture_config" in json and json["gesture_config"] is not None else None,
        )
    
    def toJSON(self) -> dict:
        return {
            "address": self.address,
            "hand": self.hand.value,
            "name": self.name,
            "touch_config": self.touch_config.toJSON() if self.touch_config else None,
            "gesture_config": self.gesture_config.toJSON() if self.gesture_config else None,
        }


import multiprocessing
import logging
import glob
import pandas as pd
import re
import sys
import time
from os import path

# resource directory - need to be referenced to exe path when used with exe bundle
# if getattr(sys, "frozen", False):
#     CLIENT_PY_PATH = sys._MEIPASS
# else:
#     CLIENT_PY_PATH = path.join(path.dirname(__file__), "../")
#     sys.path.insert(0, path.join(CLIENT_PY_PATH, "utilities"))
#     sys.path.insert(0, path.join(CLIENT_PY_PATH, "hal"))
#     sys.path.insert(0, path.join(CLIENT_PY_PATH, "middleware"))
#     sys.path.insert(0, path.join(CLIENT_PY_PATH, "data_managment/utilities"))


from .utilities import create_dir, load_info, update_dataframe, get_index, update_info
from ..hal import Ble
from ..middleware import Tactigon_Recorder
from ..models import Hand

from .models import UserData, DataCollectionConfig, HAL

def collect(user_data: UserData, hal: HAL, data_collection_config: DataCollectionConfig):
    """
    This function collect raw data from device and store in files. The function uses data_collection config file
    :return: none
    """

    print_info(data_collection_config, hal, user_data)

    # load and update info data stored in dataframes
    person_data, gestures_data, session_data, device_data = load_info(user_data, data_collection_config)    

    gestures = data_collection_config.GESTURE_NAME
    person = user_data.user_id
    session = data_collection_config.SESSION_INFO

    b_hand = data_collection_config.HAND
    b_add = hal.ADDRESS
    num_sample = hal.NUM_SAMPLE
    
    # check session name
    if get_index(session_data, "session_info", session) != "":
        print("\nSession already exists. \nPlease rename it")
        print("\n\n")
        return

    ## Update dataframe with new values from config file
    person_data = update_dataframe(person_data, "person_name", [user_data.user_id])    
    gestures_data = update_dataframe(gestures_data, "gesture_names", data_collection_config.GESTURE_NAME)
    session_data = update_dataframe(session_data, "session_info", [data_collection_config.SESSION_INFO])
    device_data = update_dataframe(device_data, "dev_id", [b_add]) 

    # pipes
    rx_sensor_pipe, tx_sensor_pipe = multiprocessing.Pipe(duplex=False)
    rx_angle_pipe, tx_angle_pipe = multiprocessing.Pipe(duplex=False)

    pro_in = Ble(b_add, b_hand)

    pro_in._sensor_tx = tx_sensor_pipe
    pro_in._angle_tx = tx_angle_pipe

    print("Waiting for Tactigon connection...")
    pro_in.start()

    while not pro_in.connected:
        pass

    # loop through one gesture at a time to collect data
    for gesture in gestures:

        # creating file name based on the information from info files       
        file_name = (
            "P"
            + get_index(person_data, "person_name", person)
            + "_G"
            + get_index(gestures_data, "gesture_names", gesture)
            + "_S"
            + get_index(session_data, "session_info", session)
            + "_D"
            + get_index(device_data, "dev_id", b_add)
            + ".csv"
        )

        gesture_path = path.join(data_collection_config.raw_data_full_path, gesture)
        file_path = path.join(gesture_path, file_name)

        pro_store = Tactigon_Recorder(b_add, file_path, gesture, num_sample, rx_sensor_pipe, rx_angle_pipe)

        # create directory for gesture if not exists
        create_dir(gesture_path)

        input("Press enter to start recording: " + gesture + " ")

        # start processes
        pro_store.start()

        # clean the pipes
        while pro_store.is_ready() == False:
            while rx_sensor_pipe.poll():
                rx_sensor_pipe.recv()
            while rx_angle_pipe.poll():
                rx_angle_pipe.recv()

        print("Press enter to stop recording: " + gesture)
        input()

        # close processes
        pro_store.terminate()

    pro_in.terminate()

    # update info files
    update_info(data_collection_config.raw_data_full_path, person_data, gestures_data, session_data, device_data)


def print_info(data_collection_config: DataCollectionConfig, hal_config: HAL, user_config: UserData):
    print("**********************************************************************")
    print("                          DATA COLLECTION                             ")
    print("**********************************************************************")
    print("             Hand: " + data_collection_config.HAND.name)
    print("             User ID: " + user_config.user_id)
    print("             Session: " + data_collection_config.SESSION_INFO)
    print("**********************************************************************")
    print("**********************************************************************")


def prepare(sessions, data_collection_config: DataCollectionConfig):
    """
    This function return a list of json data corresponding to provided sessions
    :param sessions: list of session names
    :return: list of data items ( each item {"session":session, "gesture":gesture_name, "csv":df_json})
    """
    # import json config

    list_raw_data = glob.glob(data_collection_config.raw_data_full_path + "*/*.csv")

    gesture_data = pd.read_csv(data_collection_config.raw_data_full_path + "info/gesture.csv", index_col=0)
    session_data = pd.read_csv(data_collection_config.raw_data_full_path + "info/session.csv", index_col=0)
    device_data = pd.read_csv(data_collection_config.raw_data_full_path + "info/device.csv", index_col=0)

    list_of_data = []
    empty_session = False
    for session in sessions:        
        session_id = get_index(session_data, "session_info", session)        
        if len(session_id) == 0:
            empty_session = True
        else:
            # get gestures raw data
            for ls in list_raw_data:
                # find the file name from complete path
                file_name = re.findall(r"P[\d]+_G[\d]+_S[\d]+", ls)
                if len(file_name) == 0:
                    file_name = " "
                else:
                    file_name = file_name[0]                

                    # find the session id in file name
                    if session_id == file_name.split("S")[1]:
                        
                        # read file content (csv raw data as json)
                        df_json = read_csv_as_json(ls)

                        # get gest and dev id from file name
                        gesture_id = file_name.split("_")[1][1:]
                        device_id = re.findall(r"D[\d]",ls)[0].split("D")[1]

                        # find the gesture name corresponding to gesture id
                        gesture_name = gesture_data["gesture_names"].iloc[int(gesture_id)]

                        # find device MAC corresponding to dev id
                        device_MAC = device_data["dev_id"].iloc[int(device_id)]

                        list_of_data.append(
                            {
                                "session": session,
                                "gesture": gesture_name,
                                "csv": df_json,
                                "dev_id": device_MAC
                            }
                        )

    return list_of_data, empty_session


def read_csv_as_json(path):
    """
    this function read csv and return as json
    :param path: path of the csv file
    :return: dataframe as json object
    """
    df = pd.read_csv(path, index_col=0)
    df.reset_index(drop=True, inplace=True)
    return df.to_json()


# if __name__ == "__main__":
#     collect()


__all__ = ['Client', 'collect', 'print_info', 'prepare', 'read_csv_as_json']

import sys
from os import path
import requests
import json
import datetime
import zipfile
import pickle as pickle
from .utilities import create_dir
from .models import UserData, ClientConfig, DataCollectionConfig
from .Data_Collection import prepare

from .Data_Collection import collect, print_info, prepare, read_csv_as_json


class Client:

    user_data: UserData
    cli_config: ClientConfig
    data_collection_config: DataCollectionConfig

    def __init__(self, user_data: UserData, config: ClientConfig, data_collection_config: DataCollectionConfig):

        self.user_data = user_data
        self.cli_config = config
        self.data_collection_config = data_collection_config
        
        # self.res = self.cli_config.get("RES")
        
    @property
    def url(self) -> str:
        return self.cli_config.SERVER_URL if self.cli_config.SERVER_URL[-1] == "/" else self.cli_config.SERVER_URL + "/"

    def test_connection(self):
        """
        The function make get request to teh server to get server version
        :return : server status
        :return : server version
        """

        # sending get request to the server
        res_version = requests.get(self.url + "version")
        print("SERVER IS ALIVE, SERVER VERSION: ", res_version.text)
        print("\n")

    def send_data(self):
        """
        This function send raw data to server
        :return: None
        """
        list_of_data, empty_session = prepare(self.cli_config.TRAINING_SESSIONS, self.data_collection_config)

        if not list_of_data:
            print("ERROR: Invalid Sessions list")
            return
        
        if empty_session:
            print("WARNING: One or more session has no gesture!!!")
            return

        # prepare data to send
        data = {
            "data": list_of_data,
            "auth_key": self.user_data.auth_key
        }

        # sending post request with raw data to server
        res_upload = requests.post(self.url + "upload", data=json.dumps(data))
        
        print("status: {}".format(res_upload.status_code)) 
        print("content: " + res_upload.content.decode("utf-8")) 
        print("mimetype: " + res_upload.headers['content-type']) 
        print("\n\n")

    def train_model(self):
        """
        The function make post request to server for training a new model
        """
        
        # prepare data to send
        training_request = {
            "sessions": self.cli_config.MODEL_SESSIONS,
            "model_name": self.cli_config.MODEL_NAME,
            "gestures": self.cli_config.MODEL_GESTURES,
            "split_ratio": self.cli_config.MODEL_SPLIT_RATIO,
            "auth_key": self.user_data.auth_key
        }
        
        # sending post request to the server
        res_training = requests.post(
            self.url + "training", data=json.dumps(training_request)
        )

        print("status: {}".format(res_training.status_code)) 
        print("content: " + res_training.content.decode("utf-8")) 
        print("mimetype: " + res_training.headers['content-type']) 
        print("\n\n")

    def download_model(self):
        """
        The function make post request to server for downloading a previously trained model
        """
        
        # prepare data to send
        download_request = {
            "model_name": self.cli_config.MODEL_NAME,
            "auth_key": self.user_data.auth_key
        }
                
        # sending post request to the server
        res_download = requests.post(
            self.url + "download", data=json.dumps(download_request)
        )
        
        if res_download.status_code == 200:
            # get answer and store model to files
            self.save_model(res_download.content)

            print("model save in ", self.cli_config.MODEL_DATA_PATH)
            print("\n\n")
           
        elif res_download.status_code == 201:
            print("No model downloaded")
            print("\n\n")
        
        else:
            print("status: {}".format(res_download.status_code)) 
            print("content: " + res_download.content.decode("utf-8")) 
            print("mimetype: " + res_download.headers['content-type']) 
            print("\n\n")

    def save_model(self, zip_data):
        """
        save model and info.txt in a folder named same as training id
        :param model: trained model
        :param encoder: one hot encoder object
        :return:
        """

        create_dir(self.cli_config.model_data_full_path)

        # with open(path.join(self.cli_config.model_data_full_path, "info.txt"), "w") as fl:
        #     fl.write(
        #         str(self.cli_config.__dict__)
        #         + " \ndata and time:"
        #         + datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        #     )

        # with open(path.join(self.cli_config.model_data_full_path, "info.json"), "w") as json_info:
        #     info: dict = self.cli_config.__dict__
        #     info["datetime"] = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        #     json.dump(info, json_info, indent=4)

        # save zip file
        with open(path.join(self.cli_config.model_data_full_path, "model.zip"), "wb") as f:
            f.write(zip_data)

        # unzip it
        with zipfile.ZipFile(path.join(self.cli_config.model_data_full_path, "model.zip"), "r") as zip_obj:
            zip_obj.extractall(path=self.cli_config.model_data_full_path)